//
//   摜IuWFNg DIB.h
//      by g0307 2003
//

#if !defined(AFX_DIB_H__FEE8374C_F395_45E9_9DA6_B10427CF6FA8__INCLUDED_)
#define AFX_DIB_H__FEE8374C_F395_45E9_9DA6_B10427CF6FA8__INCLUDED_

#if _MSC_VER > 1000
#pragma once
#endif // _MSC_VER > 1000

//
//  sNZ^
//
template<class T>
class pixel_t {
public:
	pixel_t(){;}
	pixel_t(T _r, T _g, T _b){
		r = _r;
		g = _g;
		b = _b;
	};//
	pixel_t(const pixel_t<float> & dest){
		r = dest.r;
		g = dest.g;
		b = dest.b;
	};//
	int norm() const{
		return r*r+g*g+b*b;
	};//
	bool operator == (const pixel_t & comp){
		return (comp.r == r && comp.g == g && comp.b == b);
	};//
	bool operator != (const pixel_t & comp){
		return !(comp.r == r && comp.g == g && comp.b == b);
	};//
	bool operator < (const pixel_t & comp){
		return norm() < comp.norm();
	};//
	pixel_t operator+ (const pixel_t<unsigned char> & dest){
		r += dest.r;
		g += dest.g;
		b += dest.b;
		return *this;
	};//
	pixel_t operator+ (const pixel_t<unsigned short> & dest){
		r += dest.r;
		g += dest.g;
		b += dest.b;
		return *this;
	};//
	pixel_t operator/ (float n){
		r /= n;
		g /= n;
		b /= n;
		return *this;
	};//
	T b;
	T g;
	T r;
};//

typedef pixel_t<unsigned char>  dib_t;
typedef pixel_t<unsigned short> rgb_t;
typedef pixel_t<float> float_t;

template <class T>
class image {
public:
	typedef	T					pixel_type;
	typedef image<T>			self_type;
	typedef self_type&			reference;
	typedef const self_type&	const_reference;
	typedef self_type*			pointer;
	typedef const self_type*	const_pointer;

	image():_bits(0),width(0),height(0),pitch(0) {}
	image(T* Bits, int Width, int Height, long Pitch):_bits(Bits),_width(Width),_height(Height),_pitch(Pitch) {
		calc_indices();
	}
	image( const_reference rhs ):_bits(0),width(0),height(0),pitch(0) {
		if ( !resize( rhs.width, rhs.height ) ) return;
		memcpy( begin(), rhs.begin(), rhs.size() );
	}
	void operator = (const_reference rhs){
		if ( !resize( rhs.width, rhs.height ) ) return;
		memcpy( begin(), rhs.begin(), rhs.size() );
	};
	virtual ~image() { clear(); }
	
	// properties
	long		size() const { return pitch * height; }

	BYTE*		get_buffer() { return _bits; }
	const BYTE*	get_buffer() const { return _bits; }
	T*			at( int x, int y ) { return reinterpret_cast<T*>(_indices[y])+x;}
	const T*	at( int x, int y ) const { return reinterpret_cast<const T*>(_indices[y])+x;}
	T*			begin() { return reinterpret_cast<T*>(_bits); }
	const T*	begin() const { return reinterpret_cast<const T*>(_bits); }
	T*			end() { return reinterpret_cast<T*>(_bits + size()); }
	const T*	end() const { return reinterpret_cast<const T*>(_bits + size()); }

	T			cubic_at(float x, float y) const{
		T ret = 0;
		int i, cx[16], cy[16];
		float dx, dy, fx, fy, w;
		
		// Cubic⊮̏ꍇPU_Kvł
		//   (ڂ͒F摜F_ pp.29Q)
		// 4ߖT
		cx[0] = int(x);cy[0] = int(y);
		cx[1] = cx[0] + 1;cy[1] = cy[0];
		cx[2] = cx[0];    cy[2] = cy[0] + 1;
		cx[3] = cx[0] + 1;cy[3] = cy[0] + 1;
		// 16ߖT
		cx[4]  = cx[0] - 1;cy[4]  = cy[0] - 1;
		cx[5]  = cx[0] - 1;cy[5]  = cy[0] + 0;
		cx[6]  = cx[0] - 1;cy[6]  = cy[0] + 1;
		cx[7]  = cx[0] - 1;cy[7]  = cy[0] + 2;
		cx[8]  = cx[0] + 0;cy[8]  = cy[0] - 1;
		cx[9]  = cx[0] + 0;cy[9]  = cy[0] + 2;
		cx[10] = cx[0] + 1;cy[10] = cy[0] - 1;
		cx[11] = cx[0] + 1;cy[11] = cy[0] + 2;
		cx[12] = cx[0] + 2;cy[12] = cy[0] - 1;
		cx[13] = cx[0] + 2;cy[13] = cy[0] + 0;
		cx[14] = cx[0] + 2;cy[14] = cy[0] + 1;
		cx[15] = cx[0] + 2;cy[15] = cy[0] + 2;
		// ݍ݌vZ
		for( i = 0; i < 16; i++){
			
			dx = fabs(cx[i] - x);
			if(dx <= 1){
				fx = 1 - 2 * dx * dx + dx * dx * dx;
			} else if(dx <= 2) {
				fx = 4 - 8 * dx + 5 * dx * dx - dx * dx * dx;
			} else fx = 0;
			dy = fabs(cy[i] - y);
			if(dy <= 1){
				fy = 1 - 2 * dy * dy + dy * dy * dy;
			} else if(dy <= 2) {
				fy = 4 - 8 * dy + 5 * dy * dy - dy * dy * dy;
			} else fy = 0;
			
			// d݂Z
			w = fx * fy;
			if(cx[i] < 0)cx[i]=-cx[i];
			if(cy[i] < 0)cy[i]=-cy[i];
			if(cx[i] >= width)cx[i]= width + width - cx[i] - 1;
			if(cy[i] >= height)cy[i]= height + height - cy[i] - 1;
			
			ret += *at(cx[i], cy[i]) * w;
			
		};//i
		
		return ret;

	};
	virtual void clear() {
		if ( is_valid() ) {
			delete [] _bits;
			_bits = NULL;
			width = height = pitch = 0;
			_indices.clear();
		}
	}

	virtual bool resize( int w, int h ) {
		clear();
		if ( w <= 0 || h <= 0 ) return false;
		width = w;
		height = h;
		pitch = ( w * sizeof(T) + 3 ) & ~3;
		_bits = new BYTE[height*pitch];
		if ( _bits == NULL ) {
			clear();
			return false;
		}
		_indices.resize( height );
		calc_indices();
		return true;
	}

	void attach(T* Bits, int Width, int Height, long Pitch) {
		claer();
		_bits = Bits;
		width = Width;
		height = Height;
		pitch = Pitch;
		BYTE* p = _bits;
		calc_indices();
	}
	void detach() {
		_bits = NULL;
		width = height= pitch = 0;
		indices.clear();
	}

	bool is_valid() const { return _bits != NULL; }

protected:
	void calc_indices() {
		_indices.resize(height);
		BYTE* p = _bits;
		for ( int i = 0; i < height; i++ ) {
			_indices[i] = p;
			p += pitch;
		}
	}

	BYTE*				_bits;
	std::vector<BYTE*>	_indices;
public:
	int					width;
	int					height;
	long				pitch;
};

typedef image<dib_t> image8_t;
typedef image<float> gray_t;
typedef image<float_t> imagef_t;

class CDIB  
{
public:
	void resize(int w, int h);
	dib_t get_cubic_pixel(float x, float y);
	void copy(CDIB &src, dib_t key);
	dib_t * at(int x, int y);
	void draw(CDC * pDC, int x, int y);
	bool read(CDC * pDC, LPSTR filename);
	CDIB();
	virtual ~CDIB();

private:

	LPBYTE lpBMP;
	BITMAPINFOHEADER BmpInfoH;
	BITMAPINFO BmpInfo;
	BITMAPFILEHEADER BmpFileH;
public:
	void write(LPSTR filename);
	void compose3(const gray_t &r, const gray_t &g, const gray_t &b);
	void decompose3(gray_t & r, gray_t & g, gray_t & b);

	int height;
	int width;
	int pitch;
};

#endif // !defined(AFX_DIB_H__FEE8374C_F395_45E9_9DA6_B10427CF6FA8__INCLUDED_)
