//**********************************************************
// s_triangle.cpp
//  by g0307 2006
//**********************************************************

#include "stdafx.h"
#include "s_triangle.h"

//******************************************************
//  from toxic
//******************************************************
bool Triangle::hit(const Ray & ray, FLOAT tmin, FLOAT tmax, HitRecord & record) const
{
     // Find vectors for two edges sharing vert0.
     const D3DXVECTOR3 edge1 = x2 - x1;
     const D3DXVECTOR3 edge2 = x3 - x1;

	 const FLOAT EPSILON = 0.000001f;
     // Begin calculating determinant - also used to calculate U parameter.
     D3DXVECTOR3 pvec, normal;
	 FLOAT t, u, v, direction;
	 D3DXVec3Cross(&pvec, &ray.v, &edge2);
 
     // If determinant is near zero, ray lies in plane of triangle.
     const FLOAT det = D3DXVec3Dot(&edge1, &pvec);
 
     // Calculate distance from vert0 to ray origin.
     const D3DXVECTOR3 tvec = ray.x - x1;
 
     const FLOAT inv_det = 1.0 / det;
 
     D3DXVECTOR3 qvec;
	 D3DXVec3Cross(&qvec, &tvec, &edge1);

     if(det > EPSILON) {
         // Calculate U parameter and test bounds.
         u = D3DXVec3Dot(&tvec, &pvec);
         if(u < 0.0 || u > det)
             return false;
 
         // Calculate V parameter and test bounds.
         v = D3DXVec3Dot(&ray.v, &qvec);
		 direction = -1;
         if(v < 0.0 || u + v > det)
             return false;
		 t = D3DXVec3Dot(&edge2, &qvec) * inv_det;

     }
     else if(det < -EPSILON) {
         // Calculate U parameter and test bounds.
         u = D3DXVec3Dot(&tvec, &pvec);
         if(u > 0.0 || u < det)
             return false; 
         // Calculate V parameter and test bounds.
         v = D3DXVec3Dot(&ray.v, &qvec);
		 direction = 1;
         if(v > 0.0 || u + v < det)
             return false;

		 t = D3DXVec3Dot(&edge2, &qvec) * inv_det;

     }
     else return false;  // ray is parallel to the plane of the triangle
 
     // Calculate t, ray intersects triangle.
     u *= inv_det;
     v *= inv_det;
	 
	 if(t >= tmin && t <= tmax){
//		 record.p = ray.x + t * ray.v;
		 record.p = x1 + edge1 * u + edge2 * v;
		 record.direction_sign = direction;
		 record.t = t;
	
		 {
//			 record.normal = 
//			 -((n1 * (1-u)+n2*u)*(1-v/(1-u))+(n3*(1-u)+n2*u) * v/(1-u)) * direction;


			 FLOAT Gu, Gv;
			 D3DXVECTOR3 phong_normal;
			 Gu = ((uv1.x * (1-u)+uv2.x*u)*(1-v/(1-u))+(uv3.x*(1-u)+uv2.x*u) * v/(1-u));
			 Gv = ((uv1.y * (1-u)+uv2.y*u)*(1-v/(1-u))+(uv3.y*(1-u)+uv2.y*u) * v/(1-u));

			 record.u = Gu;
			 record.v = Gv;


			 if(n1.x > -1000)
				 record.normal = 
				 -((n1 * (1-u)+n2*u)*(1-v/(1-u))+(n3*(1-u)+n2*u) * v/(1-u)) * direction;
			 else
				 record.normal = n * direction;
			 // ov}bvgH
			 if(pMaterial->R->fBumpmap){

				 FLOAT bw = pMaterial->R->bw;
				 FLOAT bh = pMaterial->R->bh;
				 D3DXVECTOR3 up, vp, bn, bump_n;
				 FLOAT du, dv;
				 int ref_x = int(floor(Gu * bw)), ref_y = int(floor((1-Gv) * bh));
				 ref_x = min(ref_x, bw - 2);ref_x = max(ref_x, 0);
				 ref_y = min(ref_y, bh - 2);ref_y = max(ref_y, 0);
				 du = (*pMaterial->R->map_bump->at(ref_x + 1, ref_y) - *pMaterial->R->map_bump->at(ref_x, ref_y)) / 10.0f;
				 dv = (*pMaterial->R->map_bump->at(ref_x, ref_y + 1) - *pMaterial->R->map_bump->at(ref_x, ref_y)) / 10.0f;
				 
//				 up = pu - record.normal * du;
//				 vp = pv - record.normal * dv;
//				 up = pu + n * du;
//				 vp = pv + n * dv;
//				 D3DXVec3Cross(&bn, &pu, &pv);
				 up.x = 1.0;up.y = 0.0;up.z = du;
				 vp.x = 0.0;vp.y = 1.0;vp.z = dv;
				 D3DXVec3Cross(&bn, &up, &vp);
				 D3DXVec3Normalize(&bn, &bn);

//				 record.normal = record.normal * bn.z -5*(bx * bn.x + by * bn.y);
				 bump_n = record.normal * bn.z + (pu * bn.x + pv * bn.y) * direction;
				 D3DXVec3Normalize(&bump_n, &bump_n);
				 if( -D3DXVec3Dot(&bump_n, &ray.v) < 0.3f)
					 bump_n = record.normal;

				 record.normal =  bump_n;


			 };
 
			 //@}bvgH
			 if(pMaterial->R->fNormalmap){
				 /*
				 int ref_x = int(floor(Gu * nw)), ref_y = int(floor(Gv * nh));
				 phong_normal.x = *pMaterial->R->map_nx->at(ref_x, ref_y);
				 phong_normal.y = *pMaterial->R->map_ny->at(ref_x, ref_y);
				 phong_normal.z = *pMaterial->R->map_nz->at(ref_x, ref_y);
				 phong_normal = phong_normal;
				 record.normal = -phong_normal * direction;
				 */
			 };


		 };

		 record.bx = bx;
		 record.by = by;
		 record.pMaterial = pMaterial;
		 return true;
	 }
	 return false;
};

// `
void Triangle::draw(CDC *pDC)
{
	
	CPoint dp[4];
	FLOAT scale = 20, offset = 300;
	
	dp[0] = dp[3] = CPoint(x1.x * scale + offset, x1.y * scale + offset);
	dp[1] = CPoint(x2.x * scale + offset, x2.y * scale + offset);
	dp[2] = CPoint(x3.x * scale + offset, x3.y * scale + offset);
	pDC->Polyline(dp, 4);
	
}

